import gym
from gym import error, spaces, utils
from gym.utils import seeding
import numpy as np
from os import path
import collections
import time
from utils.envs import GardenEnv, StayMoveEnv, ChainEnv, Dog1, Dog2, Dog3, TenByTenEnv
from utils.learners import Q_Learner


class AttackEnv(gym.Env):

    def __init__(self, e, learner = None, gap = 0.1, max_action = 2, heuristic = False, extended_xi = False):
        
        if e=="GardenEnv":
            env = GardenEnv()
            self.learner = Q_Learner(env.s_dim, env.a_dim, env.gamma)
        elif e=="StayMoveEnv":
            env = StayMoveEnv()
            self.learner = Q_Learner(env.s_dim, env.a_dim, env.gamma)
        elif e=="ChainEnv":
            env = ChainEnv(s_dim = 6)
            self.learner = Q_Learner(env.s_dim, env.a_dim, env.gamma)
            learner.H = 200
        elif e=="TenByTenEnv":
            env = TenByTenEnv()
            self.learner = Q_Learner(env.s_dim, env.a_dim, env.gamma)
            learner.H = 1000
    
    
        ###### Set parameters of the environment MDP
        self.env = env
        self.a_dim = 1
        ###### Set cost parameters of the attacker
        self.weight = 10.0
        self.gap = gap
        self.max_action = max_action
        self.heuristic = heuristic
        self.sorted_target_states = None
        self.extended_xi = extended_xi
        if self.extended_xi:
            self.s_dim = 2 * env.s_dim + env.a_dim + 2 * env.s_dim * env.a_dim + 1
        else:
            self.s_dim = 2 * env.s_dim + env.a_dim + env.s_dim * env.a_dim + 1
            
        self.s_dim = 2 * env.s_dim + env.a_dim + 1
            
        self.observation_space = spaces.Box(low=-1000, high=1000, shape=
                    (self.s_dim, ), dtype=np.float32)
        
        self.action_space = spaces.Box(low= - max_action, high = max_action, shape=
                    (self.a_dim, ), dtype=np.float32)
                            
    def seed(self, seed):
        np.random.seed(seed)
        
    def baseline(self):
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        if Qt_reshape[0,0] == 0 and Qt_reshape[0,1] == 0 and self.st !=0:
            if self.at == 0:
                delta = 1/self.learner.alpha * max( Qt_reshape[self.st,1] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) + 0.01
                           ,0)
            if self.at == 1:
                delta = 1/self.learner.alpha * min(Qt_reshape[self.st,0] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) - 0.01
                           ,0)
        elif self.potential(Qt_reshape)<0:
            if self.at == 0:
                delta = 1/self.learner.alpha * min( Qt_reshape[self.st,1] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) - self.gap
                           ,0)
            if self.at == 1:
                delta = 1/self.learner.alpha * max( Qt_reshape[self.st,0] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) + self.gap
                           ,0)
        else: delta = 0
        return delta
    
    def BFS(self, env, origin):
        N=env.s_dim
        queue = collections.deque([origin])
        dist = collections.defaultdict(lambda: N*N)
        dist[origin] = 0
        while queue:
            node = queue.popleft()
            d = dist[node]
            if all([s in dist for s in env.target_states]):
                return dist
                break
            for child in [env.step_sim(node,action) for action in range(env.a_dim)]:
                if d + 1 < dist[child]:
                    dist[child] = d + 1
                    queue.append(child)
                    
    def BFS_target(self, env, target, allowable_actions):
        N=env.s_dim
        queue = collections.deque([target])
        dist = collections.defaultdict(lambda: [N*N,set()])
        dist[target] = [0, set()]
        while queue:
            node = queue.popleft()
            d = dist[node][0]
#             if all([s in dist for s in range(N)]):
#                 return dist
#                 break
            for neighbor in [env.step_sim(node, action) for action in range(env.a_dim)]:
                for action in allowable_actions[neighbor]:
                    if env.step_sim(neighbor, action) == node:
                        if d + 1 <= dist[neighbor][0]:
                            dist[neighbor][0] = d+1
                            dist[neighbor][1].add(action)
                            queue.append(neighbor)    
        return dist 
    
    def greedy_baseline(self):
        if int(self.st) not in self.env.target_states:
            return 0
        else:
            Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
            if int(self.at) in self.env.attack_target[int(self.st)]:
                nontarget_actions = list(set(range(self.env.a_dim)) - set(self.env.attack_target[int(self.st)]))  
                delta = 1/self.learner.alpha * max( max(Qt_reshape[int(self.st),nontarget_actions]) - 
                            ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                            ) + self.gap
                           ,0)
            else:
                delta = 1/self.learner.alpha * min(min(Qt_reshape[int(self.st),self.env.attack_target[int(self.st)]]) - 
                            ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                            ) - self.gap
                           ,0)
        return delta
                
        
       
    def TSP_baseline(self, gap = 0.1):
        if not self.sorted_target_states:
            dist = self.BFS(self.env, self.env.s0)
            sorted_states = [x[0] for x in sorted(dist.items(), key=lambda kv: kv[1], reverse=True)]
            self.sorted_target_states = [state for state in sorted_states if state in self.env.target_states]
            self.count = 0
            self.current_target_state = self.sorted_target_states[self.count]
            self.allowable_actions = {}
            for s in range(self.env.s_dim):
                self.allowable_actions[s] = list(range(self.env.a_dim))
            for s in self.env.terminal_states:
                self.allowable_actions[s] = []
            for i in range(self.count):
                self.allowable_actions[self.sorted_target_states[i]] = self.env.attack_target[self.sorted_target_states[i]]
            self.dist2 = self.BFS_target(self.env, self.current_target_state, self.allowable_actions)
        
        if self.count>=len(self.sorted_target_states):
            return self.greedy_baseline()
        
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        

        while self.potential_individual(Qt_reshape, self.current_target_state) <= 1e-6:
            self.count += 1
            if self.count<len(self.sorted_target_states):
                self.current_target_state = self.sorted_target_states[self.count]
                self.allowable_actions = {}
                for s in range(self.env.s_dim):
                    self.allowable_actions[s] = list(range(self.env.a_dim))
                for s in self.env.terminal_states:
                    self.allowable_actions[s] = []
                for i in range(self.count):
                    self.allowable_actions[self.sorted_target_states[i]] = self.env.attack_target[self.sorted_target_states[i]]
                self.dist2 = self.BFS_target(self.env, self.current_target_state, self.allowable_actions)

            else:
                return self.greedy_baseline()
            
        
            
        
        if self.potential_individual(Qt_reshape, self.current_target_state) != 0:
            if self.st in self.sorted_target_states[0:(self.count+1)]:
                return self.greedy_baseline()
            else:
                target_actions = list(self.dist2[int(self.st)][1])
                if target_actions == []:
                    target_actions = self.allowable_actions[int(self.st)]
            
                nontarget_actions = list(set(range(self.env.a_dim)) - set(target_actions))            
                if int(self.at) in target_actions:
                    delta = 1/self.learner.alpha * max( max(Qt_reshape[int(self.st),nontarget_actions]) - 
                                ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                                ) + gap
                               ,0)
                elif int(self.at) not in target_actions:
                    delta = 1/self.learner.alpha * min(min(Qt_reshape[int(self.st),target_actions]) - 
                                ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                                ) - gap
                               ,0)
        return delta
        

    def potential_individual(self, Qt_reshape, target_state):
        key = target_state
        value = self.env.attack_target[target_state]
        return (max(max(Qt_reshape[key,list(set(range(self.env.a_dim))-set(value))]) + self.gap 
                              - max(Qt_reshape[key,value]),0))
    
    def potential(self, Qt_reshape):
        potential = 0
        for key, value in self.env.attack_target.items():
            potential -= (max(max(Qt_reshape[key,list(set(range(self.env.a_dim))-set(value))]) + self.gap 
                              - max(Qt_reshape[key,value]),0))
        return potential
            
        
    def one_hot(self, a, num_classes):
        return np.squeeze(np.eye(num_classes)[a.reshape(-1)])
    
    def attack_reward(self, delta, Qt):
        Qt_reshape = np.reshape(Qt,[self.env.s_dim,self.env.a_dim])
        return - abs(delta[0]) + self.weight * self.potential(Qt_reshape)

    def step(self,delta):
        self.stb = np.copy(self.st)
        self.atb = np.copy(self.at)
        self.stpb = np.copy(self.stp)
        self.rtb = np.copy(self.rt)
        self.Qtb = np.copy(self.Qt)
        
        if self.heuristic and self.env.s0 is not None:
            self.Qt = self.learner.update(self.st, self.at, self.rt+delta+self.TSP_baseline(gap=self.gap), self.stp, self.done)
        elif self.heuristic and self.env.s0 is None:
            self.Qt = self.learner.update(self.st, self.at, self.rt+delta+self.greedy_baseline(), self.stp, self.done)
        else:
            self.Qt = self.learner.update(self.st, self.at, self.rt+delta, self.stp, self.done)
        # Q-learning update
        if (self.learner.t % self.learner.T) == self.learner.T or self.done == True: self.env.reset()
        self.st = self.env.st
        self.at = self.learner.action(self.st)
        self.stp, self.rt, self.done, _ = self.env.step(self.at)        
        
        
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        Qt_bool = np.reshape(1.0 * np.equal(Qt_reshape,np.max(Qt_reshape, axis=1,keepdims=True)),self.env.s_dim * self.env.a_dim)  
        
        if self.extended_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                             ,Qt_bool
                            ), axis=0)
        else:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)
            
        self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt])
                            ), axis=0)
                    
        if self.heuristic and self.env.s0 is not None:
            return self.state, self.attack_reward(delta+self.TSP_baseline(gap=self.gap), self.Qt), self.learner.h==self.learner.H, {}
        elif self.heuristic and self.env.s0 is None:
            return self.state, self.attack_reward(delta+self.greedy_baseline(), self.Qt), self.learner.h==self.learner.H, {}
        else:    
            return self.state, self.attack_reward(delta, self.Qt), self.learner.h==self.learner.H, {}
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.at = np.copy(self.atb)
        self.stp = np.copy(self.stpb)
        self.rt = np.copy(self.rtb)
        self.Qt = np.copy(self.Qtb)
        self.env.rollback()
        self.learner.rollback()
        
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        Qt_bool = np.reshape(1.0 * np.equal(Qt_reshape,np.max(Qt_reshape, axis=1,keepdims=True)),self.env.s_dim * self.env.a_dim)
        
        
        if self.extended_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                             ,Qt_bool
                            ), axis=0)
        else:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)
        
        

    def reset(self):
        self.env.reset()
        self.learner.reset()
        self.st = self.env.st
        self.at = self.learner.action(self.st)
        self.stp, self.rt, self.done, _ = self.env.step(self.at)
        self.Qt = self.learner.Qt
        
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        Qt_bool = np.reshape(1.0 * np.equal(Qt_reshape,np.max(Qt_reshape, axis=1,keepdims=True)),self.env.s_dim * self.env.a_dim)
        
        if self.extended_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                             ,Qt_bool
                            ), axis=0)
        else:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)
        
        
        self.sorted_target_states = None
        
        self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt])
                            ), axis=0)
        
        return self.state

    
if __name__ == '__main__':
    e = ChainEnv()
    env = AttackEnv(e, Q_Learner(e.s_dim, e.a_dim, e.gamma), heuristic = True)
    env.reset()
    print(env.st,env.at, env.stp, env.rt, env.Qt)
    env.step(0)
    print(env.st,env.at, env.stp, env.rt, env.Qt)
    
