import gym
import tensorflow as tf
from gym import error, spaces, utils
from gym.utils import seeding
import numpy as np
from os import path



class Q_Learner(gym.Env):

    def __init__(self, s_dim, a_dim, gamma, T = 10):
        ###### Set parameters of the environment MDP
        self.s_dim = s_dim
        self.a_dim = a_dim
        self.alpha = 0.9 # learning rate for Q-learning update
        self.gamma = gamma # discounting factor in the environmental MDP
        self.epsilon = 0.1 # epsilon for epsilon greedy behavior policy
        self.H = 1000 # life span of learner
        self.T = T # environment reset frequency
      
    def action(self, st):
        Qt_reshape = np.reshape(self.Qt,[self.s_dim,self.a_dim])
        rand = np.random.rand()
        if rand<=self.epsilon:
            return np.random.choice(range(self.a_dim))
        else:
            return np.random.choice(np.argwhere(Qt_reshape[st] == np.amax(Qt_reshape[st])).flatten().tolist())
        
    def reset(self):
        self.t = 0
        self.h = 0
        self.Qt = 0.0 * np.random.randn(self.s_dim*self.a_dim)

    def update(self, st, at, rt, stp, done):
        self.Qtb = np.copy(self.Qt)
        self.t += 1
        self.h += 1
        Qt_reshape = np.reshape(self.Qt,[self.s_dim,self.a_dim])
        if not done:
            Qt_reshape[st,at] = (1-self.alpha) * Qt_reshape[st,at] + self.alpha * (rt + self.gamma * max(Qt_reshape[stp]))
        else:
            self.t = 0
            Qt_reshape[st,at] = (1-self.alpha) * Qt_reshape[st,at] + self.alpha * rt
        self.Qt = np.reshape(Qt_reshape,self.s_dim * self.a_dim)            
        return self.Qt
    
    def rollback(self):
        self.Qt = np.copy(self.Qtb)
        self.t -=1
        self.h -=1

    
if __name__ == '__main__':
    env = Q_Learner(12,4,0.9)
    print(env.reset())
    print(env.update(0,0,0,0))
    print(env.action(0))
    