import numpy as np
import torch
import gym
import argparse
import os
from utils.envs import GardenEnv, StayMoveEnv, ChainEnv, TenByTenEnv, TenByTenEnv1, TenByTenEnv2, TenByTenEnv3, Dog1, Dog2, Dog3, CartPoleEnv, Twisted_ChainEnv
from utils.AttackEnv import AttackEnv
from utils.learners import Q_Learner, Q_Learner_softmax, DQN
import TD3.utils as utils
from TD3.TD3 import TD3
import random

# Runs policy for X episodes and returns average reward
def evaluate_policy(policy, env, gamma=0.99, eval_episodes=10):
    avg_reward = 0.
    for _ in range(eval_episodes):
        obs = env.reset()
        done = False
        for h in range(1000):
            action = policy.select_action(np.array(obs))
            obs, reward, done, _ = env.step(action)
            avg_reward += gamma**h * reward
    avg_reward /= eval_episodes
    print("Sampled Action: ", action)
    print("---------------------------------------")
    print("Evaluation over %d episodes: %f" % (eval_episodes, avg_reward))
    print("---------------------------------------")
#     print(np.argmax(np.reshape(env.Qt,[env.env.s_dim,env.env.a_dim]),axis=1))
    return avg_reward


args = {}
args["policy_name"] = "TD3"               # Policy name
# args["env_name"]="StayMoveEnv"              # environment name
# args["env_name"]="GardenEnv"              # environment name
# args["env_name"]="ChainEnv"              # environment name
# args["env_name"]="TenByTenEnv"              # environment name
# args["env_name"]="TenByTenEnv1"              # environment name
# args["env_name"]="TenByTenEnv2"              # environment name
# args["env_name"]="TenByTenEnv3"              # environment name
# args["env_name"]="Twisted_ChainEnv"              # environment name


# args["env_name"]="Dog1"              # environment name
# args["env_name"]="Dog2"              # environment name
# args["env_name"]="Dog3"              # environment name
args["env_name"]="CartPoleEnv"              # environment name

args["seed"]=5                            # Sets Gym, PyTorch and Numpy seeds
args["eval_freq"]=1e4                     # How often (time steps) we evaluate
args["max_timesteps"]=5e6                 # Max time steps to run environment for
args["save_models"]="store_true"          # Whether or not models are saved
args["expl_noise"]=random.choice([0.5])   # Std of Gaussian exploration noise
args["batch_size"]=random.choice([100])                     # Batch size for both actor and critic
args["discount"]=1.0                    # Discount factor
args["policy_noise"]=random.choice([0.2]) # 0.2                  # Noise added to target policy during critic update
args["noise_clip"]=random.choice([0.5])                    # Range to clip target policy noise
args["action_l2"]=random.choice([50])
args["buffer_size"]=random.choice([1e7])
args["optimizer"]="Adam"
args["lr"] = random.choice([1e-3])
args["start_timesteps"]=random.choice([5e5])               # How many time steps purely random policy is run for
args["tau"]=random.choice([0.002])                          # Target network update rate
args["policy_freq"]=random.choice([2])                     # Frequency of delayed policy updates
args["max_action"]=random.choice([0.1])
args["lr_ratio"] = random.choice([0.5])
args["simulator"] = True
args["repeats"] = 2
args["chain_length"] = 5
args["baseline"] = False
args["shaping_weight"] = 1
args["extended_xi"] = False
args["reduced_xi"] = True

if not os.path.exists("./results"):
    os.makedirs("./results")
if args["save_models"] and not os.path.exists("./pytorch_models"):
    os.makedirs("./pytorch_models")

test_e = None
is_DQN = False

if args["env_name"]=="GardenEnv":
    e = GardenEnv()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 60)
elif args["env_name"]=="StayMoveEnv":
    e = StayMoveEnv()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma)
elif args["env_name"]=="ChainEnv":
    e = ChainEnv(s_dim = args["chain_length"])
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 5 * args["chain_length"])
    learner.H = 200
elif args["env_name"]=="Twisted_ChainEnv":
    e = Twisted_ChainEnv(s_dim = args["chain_length"])
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 5 * args["chain_length"])
    learner.H = 1000
elif args["env_name"]=="TenByTenEnv":
    e = TenByTenEnv()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 30)
    learner.H = 1000
    args["repeats"] = 2
elif args["env_name"]=="TenByTenEnv1":
    e = TenByTenEnv1()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 100)
    learner.H = 1000
    args["repeats"] = 2
elif args["env_name"]=="TenByTenEnv2":
    e = TenByTenEnv2()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 100)
    learner.H = 1000
    args["repeats"] = 2
elif args["env_name"]=="TenByTenEnv3":
    e = TenByTenEnv3()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma, T = 100)
    learner.H = 1000
    args["repeats"] = 1

elif args["env_name"]=="Dog1":    
    e = Dog1()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma)
    learner.H = 100
elif args["env_name"]=="Dog2":    
    e = Dog2()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma)
    learner.H = 200
elif args["env_name"]=="Dog3":    
    e = Dog3()
    learner = Q_Learner(e.s_dim, e.a_dim, e.gamma)
    learner.H = 500
    
elif args["env_name"]== 'CartPoleEnv':
    e = CartPoleEnv(reward = 1.0)
    test_e = CartPoleEnv(reward = 0.1)
    learner = DQN(state_dim =e.s_dim, action_dim = e.a_dim, batch_size = 1,
                  learning_rate = 0.01, tau = 0.1, iterations = 1, target_freq = 1, freeze = True, buffer_size = 1)
    is_DQN = True
    args["repeats"] = 1

    
    
env = AttackEnv(e, learner, gap = 0.1, shaping_weight = args["shaping_weight"], heuristic = args["baseline"], extended_xi = args["extended_xi"], DQN = is_DQN, test_env = test_e, reduced_xi = args["reduced_xi"], max_action = args["max_action"])

if args["env_name"]== 'CartPoleEnv':
    env.weight = 1

file_name = "%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s_%s" % (args["env_name"],
                                                     args["policy_name"],
                                                  str(args["lr_ratio"]),
                                                  str(args["start_timesteps"]),
                                                  str(args["tau"]),
                                                  str(args["policy_freq"]),
                                                  str(args["lr"]),
                                                  str(args["action_l2"]),
                                                  str(args["noise_clip"]),
                                                  str(args["policy_noise"]),
                                                  str(args["expl_noise"]),
                                                  str(args["max_action"]),
                                                  str(args["repeats"]),   
                                                  str(args["chain_length"]),
                                                  str(args["baseline"]),
                                                  str(args["shaping_weight"]),
                                                  str(args["extended_xi"]),
                                                  str(args["reduced_xi"]),
                                                  str(args["discount"])
                                                      )
print("---------------------------------------")
print("Settings: %s" % (file_name))
print("---------------------------------------")


# Set seeds
env.seed(args["seed"])
torch.manual_seed(args["seed"])
np.random.seed(args["seed"])

state_dim = env.s_dim
action_dim = env.a_dim
max_action = args["max_action"]

# Initialize policy
policy = TD3(state_dim, action_dim, max_action, args["optimizer"], args["lr"], args["lr_ratio"])
# policy.load("CartPoleEnv_TD3_0.5_500000.0_0.002_2_0.001_50_0.5_0.2_0.5_2_1_6_False_1_False",
#                                    "./pytorch_models") 

replay_buffer = utils.ReplayBuffer(max_size=args["buffer_size"])

# Evaluate untrained policy
evaluations = [evaluate_policy(policy, env, args["discount"])] 

total_timesteps = 0
timesteps_since_eval = 0
episode_num = 0
done = True 
best_reward = -float('Inf')
while total_timesteps < args["max_timesteps"]:

    if done: 

        if total_timesteps != 0: 
            print(("Total T: %d Episode Num: %d Episode T: %d Reward: %f") % (total_timesteps, episode_num, episode_timesteps, episode_reward))
            policy.train(replay_buffer, episode_timesteps, args["batch_size"], args["discount"], args["tau"], args["policy_noise"], args["noise_clip"], args["policy_freq"], args["action_l2"])

        # Evaluate episode
        if timesteps_since_eval >= args["eval_freq"]:
            timesteps_since_eval %= args["eval_freq"]
            evaluations.append(evaluate_policy(policy, env, args["discount"]))
            np.save("./results/%s" % (file_name), evaluations)
            if evaluations[-1]>best_reward:
                if os.path.exists("./pytorch_models/"+file_name+"_"+str(-best_reward)+"_actor.pth"):
                    os.remove("./pytorch_models/"+file_name+"_"+str(-best_reward)+"_actor.pth")
                    os.remove("./pytorch_models/"+file_name+"_"+str(-best_reward)+"_critic.pth")
                    os.remove("./pytorch_models/"+file_name+"_"+str(-best_reward)+"_actor_target.pth")
                    os.remove("./pytorch_models/"+file_name+"_"+str(-best_reward)+"_critic_target.pth")
                best_reward = evaluations[-1]
                if args["save_models"]: 
                    policy.save(file_name+"_"+str(-best_reward), directory="./pytorch_models")
                    
            

        # Reset environment
        obs = env.reset()
        done = False
        episode_reward = 0
        episode_timesteps = 0
        episode_num += 1 

    # Select action randomly or according to policy
    if total_timesteps < args["start_timesteps"] and args["baseline"] == False:
        action = (np.random.normal(0, max_action, size=action_dim)).clip(-max_action, max_action)
    else:
        action = policy.select_action(np.array(obs))
        if args["expl_noise"] != 0: 
            action = (action + np.random.normal(0, args["expl_noise"], size=action_dim)).clip(-max_action, max_action)

    # Perform action
    new_obs = []
    if args["simulator"]:
        for repeat in range(args["repeats"]-1):
            new_ob, reward, done, _ = env.step(action)
            done_bool = 0 if episode_timesteps + 1 == learner.H else float(done)
            new_obs.append(new_ob)
            env.rollback()
        
    new_ob, reward, done, _ = env.step(action) 
    done_bool = 0 if episode_timesteps + 1 == learner.H else float(done)
    new_obs.append(new_ob)
    episode_reward += reward

    # Store data in replay buffer
    replay_buffer.add((obs, new_obs, action, reward, done_bool))

    obs = new_ob

    episode_timesteps += 1
    total_timesteps += 1
    timesteps_since_eval += 1

# Final evaluation 
evaluations.append(evaluate_policy(policy, env, args["discount"])) 
if args["save_models"]: policy.save("%s" % (file_name), directory="./pytorch_models")
np.save("./results/%s" % (file_name), evaluations)  