import gym
import tensorflow as tf
from gym import error, spaces, utils
from gym.utils import seeding
import numpy as np
from os import path
import collections
import time
import copy as cp
import cvxpy as cvx
import utils.attack_def as ad

# from envs import GardenEnv, StayMoveEnv, ChainEnv
# from learners import Q_Learner


class AttackEnv(gym.Env):

    def __init__(self, env, learner, gap = 0.1, shaping_weight = 1, heuristic = False, extended_xi = False, randomized_xi_0 = False, DQN = False, test_env = None, reduced_xi = False, max_action = None, shaping = False):
        ###### Set parameters of the environment MDP
        self.env = env
        self.test_env = test_env
        self.learner = learner
        self.DQN = DQN
        self.a_dim = 1
        ###### Set cost parameters of the attacker
        self.weight = 10.0
        self.gap = gap
        self.heuristic = heuristic
        self.sorted_target_states = None
        self.extended_xi = extended_xi
        self.reduced_xi = reduced_xi
        self.max_action = max_action
        self.shaping = shaping
        self.delta3_attack_table = self.delta3_attack(self.env.s_dim, self.env.gamma, self.max_action)
        if self.extended_xi:
            self.s_dim = 2 * env.s_dim + env.a_dim + 2 * env.s_dim * env.a_dim + 1
        elif self.reduced_xi:
            self.s_dim = 2 * env.s_dim + env.a_dim
        else:
            self.s_dim = 2 * env.s_dim + env.a_dim + env.s_dim * env.a_dim + 1
        if self.DQN:
            self.s_dim = 2 * env.s_dim + env.a_dim + 1 + self.learner.param_dim
            
        self.visits = np.zeros([self.env.s_dim, self.env.a_dim])
        
        self.shaping_weight = shaping_weight
        self.randomized_xi_0 = randomized_xi_0
        
    def seed(self, seed):
        np.random.seed(seed)
    
    def delta3_attack(self, L, gamma , Delta):
        tar = {0:[0]}
        for s in range(1,L-1):
            tar[s] = [1]

        S = list(range(L))
        A = [0, 1]
        learner = ad.TCE(gamma)
        rewards = -0.1 * np.ones(L) # grid world rewards
        rewards[-1] += 1    
        D0 = [(s,a,rewards[ max(s - 1,0) if a==0 else min(s + 1, L-1)], 
               max(s - 1,0) if a==0 else min(s + 1, L-1)) for s in range(L-1) for a in A]
        D0.append((L-1,0,0,L-1))
        D0.append((L-1,1,0,L-1))

        r0 = np.array([D0[t][2] for t in range(len(D0))])

        # learn on clean data D0
        learner.estimate(D0, S, A)
        Q0,Q0_traj = learner.value_iteration()

        original_policy = np.argmax(Q0,axis=1)

        max_diff = -float('inf')
        for s in tar:
            M_non = max([Q0[s,a] for a in set(A).difference(set(tar[s]))])
            M_tar = max([Q0[s,a] for a in set(tar[s])])
            max_diff = max(max_diff, M_non - M_tar)
        Delta_3 = (1 + gamma) * max_diff / 2

        Qp = cp.deepcopy(Q0)

        for s in tar:
            for a in A:
                if a in tar[s]:
                    Qp[s,a] += (Delta + Delta_3) / (2 * (1 + gamma))
                else:
                    Qp[s,a] -= (Delta + Delta_3) / (2 * (1 + gamma))

        Rp = np.array([[0] * len(A) for i in range(len(S))])
        Subtract = np.array([[0.0] * len(A) for i in range(len(S))])
        for s in S:
            for a in A:
                Suc = learner.Phat[(s,a)]
                Sum = 0
                for sp in S:
                    Sum += Suc[sp] * max([Qp[sp,ap] for ap in range(len(A))])
                Subtract[s,a] = gamma * Sum

        Rp = Qp - Subtract
        return Rp - learner.Rhat
    
    def baseline(self):
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        if Qt_reshape[0,0] == 0 and Qt_reshape[0,1] == 0 and self.st !=0:
            if self.at == 0:
                delta = 1/self.learner.alpha * max( Qt_reshape[self.st,1] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) + 0.01
                           ,0)
            if self.at == 1:
                delta = 1/self.learner.alpha * min(Qt_reshape[self.st,0] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) - 0.01
                           ,0)
        elif self.potential(Qt_reshape)<0:
            if self.at == 0:
                delta = 1/self.learner.alpha * min( Qt_reshape[self.st,1] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) - self.gap
                           ,0)
            if self.at == 1:
                delta = 1/self.learner.alpha * max( Qt_reshape[self.st,0] - 
                            (
                                (1-self.learner.alpha)*Qt_reshape[self.st,self.at] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[self.stp]))
                            ) + self.gap
                           ,0)
        else: delta = 0
        return delta
    
    def BFS(self, env, origin):
        N=env.s_dim
        queue = collections.deque([origin])
        dist = collections.defaultdict(lambda: N*N)
        dist[origin] = 0
        while queue:
            node = queue.popleft()
            d = dist[node]
            if all([s in dist for s in env.target_states]):
                return dist
                break
            for child in [env.step_sim(node,action) for action in range(env.a_dim)]:
                if d + 1 < dist[child]:
                    dist[child] = d + 1
                    queue.append(child)
                    
    def BFS_target(self, env, target, allowable_actions):
        N=env.s_dim
        queue = collections.deque([target])
        dist = collections.defaultdict(lambda: [N*N,set()])
        dist[target] = [0, set()]
        while queue:
            node = queue.popleft()
            d = dist[node][0]
#             if all([s in dist for s in range(N)]):
#                 return dist
#                 break
            for neighbor in [env.step_sim(node, action) for action in range(env.a_dim)]:
                for action in allowable_actions[neighbor]:
                    if env.step_sim(neighbor, action) == node:
                        if d + 1 <= dist[neighbor][0]:
                            dist[neighbor][0] = d+1
                            dist[neighbor][1].add(action)
                            queue.append(neighbor)    
        return dist 
    
    def greedy_baseline(self):
        if int(self.st) not in self.env.target_states:
            return 0
        else:
            Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
            if int(self.at) in self.env.attack_target[int(self.st)]:
                nontarget_actions = list(set(range(self.env.a_dim)) - set(self.env.attack_target[int(self.st)]))  
                delta = 1/self.learner.alpha * max( max(Qt_reshape[int(self.st),nontarget_actions]) - 
                            ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                            ) + self.gap
                           ,0)
            else:
                delta = 1/self.learner.alpha * min(min(Qt_reshape[int(self.st),self.env.attack_target[int(self.st)]]) - 
                            ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                            ) - self.gap
                           ,0)
        if self.max_action is None:
            return delta
        else:
            return max(min(delta,self.max_action), -self.max_action)
                
    def shaping_baseline(self):
        deltas = np.array([[ 0.13338491, -0.16579225],
       [-0.04113325,  0.00872591],
       [-0.16579225,  0.00872591],
       [-0.16579225,  0.08725908],
       [ 0.        ,  0.        ]])
        return deltas[int(self.st),int(self.at)]
       
    def TSP_baseline(self, gap = 0.1):
        if not self.sorted_target_states:
            dist = self.BFS(self.env, self.env.s0)
            sorted_states = [x[0] for x in sorted(dist.items(), key=lambda kv: kv[1], reverse=True)]
            self.sorted_target_states = [state for state in sorted_states if state in self.env.target_states]
#             self.sorted_target_states = [9,8,7,6,5,4,3,2,1,0,10,20,30,40,50,60,70,80]
            self.count = 0
            self.current_target_state = self.sorted_target_states[self.count]
            self.allowable_actions = {}
            for s in range(self.env.s_dim):
                self.allowable_actions[s] = list(range(self.env.a_dim))
            for s in self.env.terminal_states:
                self.allowable_actions[s] = []
            for i in range(self.count):
                self.allowable_actions[self.sorted_target_states[i]] = self.env.attack_target[self.sorted_target_states[i]]
            self.dist2 = self.BFS_target(self.env, self.current_target_state, self.allowable_actions)
        
        if self.count>=len(self.sorted_target_states):
            return self.greedy_baseline()
        
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        

        while self.potential_individual(Qt_reshape, self.current_target_state) <= 1e-6:
            self.count += 1
            if self.count<len(self.sorted_target_states):
                self.current_target_state = self.sorted_target_states[self.count]
                self.allowable_actions = {}
                for s in range(self.env.s_dim):
                    self.allowable_actions[s] = list(range(self.env.a_dim))
                for s in self.env.terminal_states:
                    self.allowable_actions[s] = []
                for i in range(self.count):
                    self.allowable_actions[self.sorted_target_states[i]] = self.env.attack_target[self.sorted_target_states[i]]
                self.dist2 = self.BFS_target(self.env, self.current_target_state, self.allowable_actions)

            else:
                return self.greedy_baseline()
            
        
            
        
        if self.potential_individual(Qt_reshape, self.current_target_state) != 0:
            if self.st in self.sorted_target_states[0:(self.count+1)]:
                return self.greedy_baseline()
            else:
                target_actions = list(self.dist2[int(self.st)][1])
                if target_actions == []:
                    target_actions = self.allowable_actions[int(self.st)]
            
                nontarget_actions = list(set(range(self.env.a_dim)) - set(target_actions))            
                if int(self.at) in target_actions:
                    delta = 1/self.learner.alpha * max( max(Qt_reshape[int(self.st),nontarget_actions]) - 
                                ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                                ) + gap
                               ,0)
                elif int(self.at) not in target_actions:
                    delta = 1/self.learner.alpha * min(min(Qt_reshape[int(self.st),target_actions]) - 
                                ((1-self.learner.alpha)*Qt_reshape[int(self.st),int(self.at)] + self.learner.alpha * (self.rt +self.env.gamma * max(Qt_reshape[int(self.stp)]))
                                ) - gap
                               ,0)
        if self.max_action is None:
            return delta
        else:
            return max(min(delta,self.max_action), -self.max_action)
        
    def DQN_baseline(self):
        return -1.1
        
    def potential_individual(self, Qt_reshape, target_state):
        key = target_state
        value = self.env.attack_target[target_state]
        return (max(max(Qt_reshape[key,list(set(range(self.env.a_dim))-set(value))]) + self.gap 
                              - max(Qt_reshape[key,value]),0))
    
    def potential(self, Qt_reshape):
        potential = 0
        for key, value in self.env.attack_target.items():
            potential -= (max(max(Qt_reshape[key,list(set(range(self.env.a_dim))-set(value))]) + self.gap 
                              - max(Qt_reshape[key,value]),0))
        return potential
            
        
    def one_hot(self, a, num_classes):
        return np.squeeze(np.eye(num_classes)[a.reshape(-1)])
    
    def attack_reward(self, delta, Qt):
        Qt_reshape = np.reshape(Qt,[self.env.s_dim,self.env.a_dim])
        return self.potential(Qt_reshape)

#         return - abs(delta) + self.weight * self.potential(Qt_reshape)
    
    def evaluate(self, env, learner, T = 1):
        start = time.time()
        
        total_reward = 0
        for j in range(T):
            obs = env.reset()
            done = False
            for i in range(200):
                action = learner.action(obs, epsilon = 0)
                obs, r, done, _ = env.step(action)
                total_reward += r
                if done: break
        end = time.time()
#         print('DQN evaluate elapse time: ', end - start)
        return total_reward/T
    
    def DQN_attack_reward(self, delta):
        return - abs(delta) - self.weight * self.evaluate(self.test_env, self.learner)
    
    def step(self,delta):
        self.stb = np.copy(self.st)
        self.atb = np.copy(self.at)
        self.stpb = np.copy(self.stp)
        self.rtb = np.copy(self.rt)
        self.Qtb = np.copy(self.Qt)
        self.doneb = self.done
#         if self.DQN:
#             start = time.time()
#             self.Qt = self.learner.update(self.st, self.at, self.rt+delta[0]+self.DQN_baseline(), self.stp, self.done, False)
#             end = time.time()
#             print('DQN update elapse time: ', end - start)
        if self.shaping:
            real_delta = (delta+self.shaping_baseline()).clip(-self.max_action, self.max_action)
        elif self.reduced_xi:
            real_delta = (delta+self.delta3_attack_table[int(self.st),int(self.at)]).clip(-self.max_action, self.max_action)
        elif self.heuristic and self.env.s0 is not None:
            real_delta = (delta+self.TSP_baseline(gap=self.gap)).clip(-self.max_action, self.max_action)
        elif self.heuristic and self.env.s0 is None:
            real_delta = (delta+self.greedy_baseline()).clip(-self.max_action, self.max_action)
        else:
            real_delta = (delta).clip(-self.max_action, self.max_action)
            
        self.Qt = self.learner.update(self.st, self.at, self.rt+real_delta, self.stp, self.done)
            
            
            
        # Q-learning update
        if self.learner.t == self.learner.T or self.done == True: 
            self.st = self.env.reset()
            self.learner.t = 0
        else:
            self.st = self.stp

        
        self.at = self.learner.action(self.st)
        self.stp, self.rt, self.done, _ = self.env.step(self.at)
        
        if self.DQN:
            self.state = np.concatenate((self.st,
                                         self.one_hot(np.array(self.at),self.env.a_dim),
                                         self.stp,
                                         np.array([self.rt]),
                                         self.Qt
                                        ), axis=0)
            return self.state, self.DQN_attack_reward(delta[0]+self.DQN_baseline()), self.learner.h==self.learner.H, {}
            
            
            
            
        if self.extended_xi:
            Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
            Qt_bool = np.reshape(1.0 * np.equal(Qt_reshape,np.max(Qt_reshape, axis=1,keepdims=True)),self.env.s_dim * self.env.a_dim) 
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                             ,Qt_bool
                            ), axis=0)
        elif self.reduced_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim)
                            ), axis=0)
        else:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)
            
            
        self.visits[self.st,self.at] += 1
        return self.state, self.attack_reward(real_delta, self.Qt), self.learner.h==self.learner.H, {}
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.at = np.copy(self.atb)
        self.stp = np.copy(self.stpb)
        self.rt = np.copy(self.rtb)
        self.Qt = np.copy(self.Qtb)
        self.done = self.doneb
        self.env.rollback()
        self.learner.rollback()
        
        Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
        Qt_bool = np.reshape(1.0 * np.equal(Qt_reshape,np.max(Qt_reshape, axis=1,keepdims=True)),self.env.s_dim * self.env.a_dim)
        
        
        if self.extended_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                             ,Qt_bool
                            ), axis=0)
        elif self.reduced_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim)
                            ), axis=0)
        else:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)
        
        

    def reset(self):
        self.st = self.env.reset()
        self.Qt = self.learner.reset()
        if self.randomized_xi_0:
            self.env.st = np.random.choice(list(range(self.env.s_dim)))
            self.learner.Qt = 1.0 * np.random.randn(self.learner.s_dim*self.learner.a_dim)
        self.at = self.learner.action(self.st)
        self.stp, self.rt, self.done, _ = self.env.step(self.at)
        
        if self.DQN:
            self.state = np.concatenate((self.st,
                                         self.one_hot(np.array(self.at),self.env.a_dim),
                                         self.stp,
                                         np.array([self.rt]),
                                         self.Qt
                                        ), axis=0)
            return self.state
        
        if self.extended_xi:
            Qt_reshape = np.reshape(self.Qt,[self.env.s_dim,self.env.a_dim])
            Qt_bool = np.reshape(1.0 * np.equal(Qt_reshape,np.max(Qt_reshape, axis=1,keepdims=True)),self.env.s_dim * self.env.a_dim)

            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                             ,Qt_bool
                            ), axis=0)
        elif self.reduced_xi:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim)
                            ), axis=0)
            
        else:
            self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)
        
        
        self.sorted_target_states = None
        
        return self.state
    
    def randomize(self):
        self.st = self.env.randomize()
        self.Qt = self.learner.randomize()
        self.at = self.learner.action(self.st)
        self.stp, self.rt, self.done, _ = self.env.step(self.at)
        
        self.state = np.concatenate((self.one_hot(np.array(self.st),self.env.s_dim),
                             self.one_hot(np.array(self.at),self.env.a_dim),
                             self.one_hot(np.array(self.stp),self.env.s_dim),
                             np.array([self.rt]),
                             self.Qt
                            ), axis=0)  
        return self.state

    
if __name__ == '__main__':
    e = ChainEnv()
    env = AttackEnv(e, Q_Learner(e.s_dim, e.a_dim, e.gamma), heuristic = True)
    env.reset()
    print(env.st,env.at, env.stp, env.rt, env.Qt)
    env.step(0)
    print(env.st,env.at, env.stp, env.rt, env.Qt)
    
